# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Processing Deployed Changes
"""

from __future__ import unicode_literals

import os
import os.path
import re
import csv
import logging
import datetime
from decimal import Decimal

import pytz

from sqlalchemy.orm.exc import NoResultFound

from locsms.db import Session as SmsSession
from locsms.db import model as locsms

from rattail.core import Object
from rattail.gpc import GPC
from rattail.db import Session
from rattail.db import model
from rattail.db.api.products import get_product_by_upc
from rattail import enum
from rattail.util import load_object
from rattail.filemon import Action
from rattail.files import overwriting_move
from rattail.time import localtime


change_pattern_generic = re.compile(r'^(?P<batch_id>\d{8})_(?P<batch_name>.+)$')
change_pattern = re.compile(r'^\d{8}_(?P<batch_name>.+)\.csv$', re.IGNORECASE)

log = logging.getLogger(__name__)


class ChangesProcessor(object):
    """
    Base class and default implementation for processing changes which have
    been deployed from SMS.
    """

    def __init__(self, config):
        self.config = config

    def process_all_changes(self, session, changes):
        log.debug("processing changes: {0}".format(repr(changes.keys())))

        for change_name, func in self.relevant_changes():
            log.debug("looking for change file: {0}".format(repr(change_name)))

            if change_name in changes:
                csv_path = changes.pop(change_name)
                log.debug("processing change file: {0}".format(repr(csv_path)))

                csv_file = open(csv_path, 'rb')
                reader = csv.DictReader(csv_file, quotechar=b"'")
                for row in reader:
                    log.debug("calling func {0} with row: {1}".format(
                            repr(func.__name__), repr(row)))
                    func(row, session)
                    session.flush()
                csv_file.close()

            else:
                log.debug("no change file found: {0}".format(repr(change_name)))

        for csv_path in changes.itervalues():
            log.warning("ignoring change file: {0}".format(repr(csv_path)))

    def relevant_changes(self):

        # Process deletions.
        yield 'CLK_DEL', self.process_clk_del
        yield 'CLT_DEL', self.process_clt_del
        yield 'COST_DEL', self.process_cost_del
        yield 'PRICE_DEL', self.process_price_del
        yield 'ALT_DEL', self.process_alt_del
        yield 'POS_DEL', self.process_pos_del
        yield 'OBJ_DEL', self.process_obj_del
        yield 'VENDOR_DEL', self.process_vendor_del

        # Process creations and updates.
        yield 'VENDOR_CHG', self.process_vendor_chg
        yield 'OBJ_CHG', self.process_obj_chg
        yield 'POS_CHG', self.process_pos_chg
        yield 'ALT_CHG', self.process_alt_chg
        yield 'PRICE_CHG', self.process_price_chg
        yield 'COST_CHG', self.process_cost_chg
        yield 'CLT_CHG', self.process_clt_chg
        yield 'CLK_CHG', self.process_clk_chg

    def date(self, val):
        return datetime.datetime.strptime(val, '%m/%d/%Y').date() if val else None

    def time(self, val):
        return datetime.datetime.strptime(val, '%H:%M').time() if val else None

    def decimal(self, val):
        return Decimal(val.strip()) if val.strip() else None

    def int_(self, val):
        return int(val) if val else None

    def utc_timestamp(self, date, time):
        """
        Return a datetime which combines the two parameters, and is localized
        to the UTC time zone, but which is still naive.  The idea here is to
        obtain a value which may be compared directly to those which are stored
        within the Rattail database, since those are naive UTC as well.
        """
        if date is None:
            return None
        if time is None:
            time = datetime.time(0) # assume midnight
        time = datetime.datetime.combine(date, time)
        time = localtime(self.config, time, 'rattail.sw.locsms')
        time = time.astimezone(pytz.utc)
        return time.replace(tzinfo=None)

    def get_product(self, upc, session):
        upc = GPC(upc, calc_check_digit='upc')
        q = session.query(model.Product)
        q = q.filter(model.Product.upc == upc)
        prod = q.first()
        if not prod:
            prod = model.Product(upc=upc)
            session.add(prod)
        return prod

    def get_price_object(self, row):

        p = Object()
        p.F126 = self.int_(row['F126'])

        # Regular Price
        p.F35 = self.date(row['F35'])
        p.F36 = self.time(row['F36'])
        p.F129 = self.date(row['F129'])
        p.F130 = self.time(row['F130'])
        p.F30 = self.decimal(row['F30'])
        p.F31 = self.int_(row['F31'])
        p.F140 = self.decimal(row['F140'])
        p.F142 = self.int_(row['F142'])
        p.F63 = self.decimal(row['F63'])
        p.F62 = self.int_(row['F62'])
        p.F111 = self.decimal(row['F111'])
        p.F112 = self.decimal(row['F112'])

        # TPR
        p.F183 = self.date(row['F183'])
        p.F184 = self.date(row['F184'])
        p.F181 = self.decimal(row['F181'])
        p.F182 = self.int_(row['F182'])
        p.F1186 = self.decimal(row['F1186'])
        p.F1187 = self.int_(row['F1187'])
        p.F1188 = self.decimal(row['F1188'])
        p.F1189 = self.int_(row['F1189'])
        p.F1218 = self.decimal(row['F1218'])
        p.F1219 = self.decimal(row['F1219'])

        # Sale
        p.F137 = self.date(row['F137'])
        p.F144 = self.time(row['F144'])
        p.F138 = self.date(row['F138'])
        p.F145 = self.time(row['F145'])
        p.F136 = self.decimal(row['F136'])
        p.F135 = self.int_(row['F135'])
        p.F139 = self.decimal(row['F139'])
        p.F143 = self.int_(row['F143'])
        p.F148 = self.decimal(row['F148'])
        p.F147 = self.int_(row['F147'])
        p.F1220 = self.decimal(row['F1220'])
        p.F1221 = self.decimal(row['F1221'])
        
        return p

    def utcnow(self):
        """
        Return the current naive UTC date and time.  (This is abstracted mostly
        for the convenience of tests.)
        """
        return datetime.datetime.utcnow()

    def price_is_current(self, price):
        """
        Determine if the given :class:`rattail.db.model.ProductPrice` instance
        should be considered "current", i.e. if the current date and time would
        fall "inside" the price's start and/or end datetimes.
        """
        now = self.utcnow()
        if price.starts and price.ends:
            return price.starts <= now <= price.ends
        if price.starts:
            return price.starts <= now
        if price.ends:
            return now <= price.ends
        return True

    def copy_reg_price(self, sms_price, price):
        price.ends = self.utc_timestamp(sms_price.F129, sms_price.F130)
        price.price = sms_price.F30
        price.multiple = sms_price.F31 or 1
        if sms_price.F140:
            price.pack_price = sms_price.F140
            price.pack_multiple = sms_price.F142 or 1
        else:
            price.pack_price = None
            price.pack_multiple = None

    def copy_tpr_price(self, sms_price, price):
        price.ends = self.utc_timestamp(sms_price.F184, None)
        price.price = sms_price.F181
        price.multiple = sms_price.F182 or 1
        if sms_price.F1186:
            price.pack_price = sms_price.F1186
            price.pack_multiple = sms_price.F1187 or 1
        else:
            price.pack_price = None
            price.pack_multiple = None

    def copy_sale_price(self, sms_price, price):
        price.ends = self.utc_timestamp(sms_price.F138, sms_price.F145)
        price.price = sms_price.F136
        price.multiple = sms_price.F135 or 1
        if sms_price.F139:
            price.pack_price = sms_price.F139
            price.pack_multiple = sms_price.F143 or 1
        else:
            price.pack_price = None
            price.pack_multiple = None

        # Apply discount if one is present.
        if sms_price.F1221:
            if price.price:
                price.price = (price.price * (100 - sms_price.F1221) / 100)\
                    .quantize(Decimal('0.01'))
            if price.pack_price:
                price.pack_price = (price.pack_price * (100 - sms_price.F1221) / 100)\
                    .quantize(Decimal('0.01'))

    def process_vendor_chg(self, row, session):
        q = session.query(model.Vendor)
        q = q.filter(model.Vendor.id == row['F27'])
        vendor = q.first()
        if vendor:
            while vendor.phones:
                vendor.phones.pop()
        else:
            vendor = model.Vendor(id=row['F27'])
            session.add(vendor)

        vendor.name = row['F334']
        vendor.special_discount = self.int_(row['F347'])

        if row['F341']:
            vendor.add_phone_number(row['F341'])
        if row['F1890']:
            vendor.add_phone_number(row['F1890'])
        if row['F342']:
            vendor.add_phone_number(row['F342'], type='Fax')
        if row['F1891']:
            vendor.add_phone_number(row['F1891'], type='Fax')

        def get_contact(name):
            for contact in vendor._contacts:
                if contact.person.display_name == name:
                    return contact
            return None

        def set_contact(contact, preference):
            if contact.preference > preference:
                contact = vendor._contacts.pop(contact.preference - 1)
                vendor._contacts.insert(preference - 1, contact)

        next_preference = 1

        if row['F335']:
            contact = get_contact(row['F335'])
            if not contact:
                person = model.Person(display_name=row['F335'])
                contact = model.VendorContact(person=person)
                vendor._contacts.append(contact)
            set_contact(contact, next_preference)
            next_preference += 1

        if row['F1889']:
            contact = get_contact(row['F1889'])
            if not contact:
                person = model.Person(display_name=row['F1889'])
                contact = model.VendorContact(person=person)
                vendor._contacts.append(contact)
            set_contact(contact, next_preference)
            next_preference += 1

        while len(vendor._contacts) > next_preference - 1:
            contact = vendor._contacts.pop()
            session.delete(contact)

        email = row['F2603'].strip() or None
        if email:
            if vendor.email:
                if vendor.email.address != email:
                    vendor.email.address = email
            else:
                vendor.add_email_address(email)
        elif vendor.emails:
            del vendor.emails[:]

        return vendor

    def process_vendor_del(self, row, session):
        q = session.query(model.Vendor)
        q = q.filter(model.Vendor.id == row['F27'])
        vendor = q.first()
        if not vendor:
            return

        # Remove associated ProductCost records.
        q = session.query(model.ProductCost)
        q = q.filter(model.ProductCost.vendor == vendor)
        q.delete(synchronize_session=False)

        session.delete(vendor)

    def process_obj_chg(self, row, session):
        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return

        product = self.get_product(upc, session)
        product.description = row['F29']
        product.size = row['F22']

        brand = None
        brand_name = row['F155'].strip()
        if brand_name:
            brand = session.query(model.Brand).filter_by(name=brand_name).first()
            if not brand:
                brand = model.Brand(name=brand_name)
                session.add(brand)
        product.brand = brand

        family = None
        F16 = self.int_(row[u'F16'])
        if F16:
            family = session.query(model.Family)\
                .filter(model.Family.code == F16)\
                .first()
            if not family:
                family = model.Family(code=F16)
                session.add(family)
        product.family = family

        report_code = None
        F18 = self.int_(row[u'F18'])
        if F18:
            report_code = session.query(model.ReportCode)\
                .filter(model.ReportCode.code == F18)\
                .first()
            if not report_code:
                report_code = model.ReportCode(code=F18)
                session.add(report_code)
        product.report_code = report_code

        product.unit_size = self.decimal(row['F1744'])

        product.deleted = False
        return product

    def get_upc(self, row):
        """
        Return the UPC (F01) field for the row, as a ``GPC`` object, or ``None``.
        """
        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return
        return GPC(upc, calc_check_digit='upc')

    def process_obj_del(self, row, session):
        upc = self.get_upc(row)
        if not upc:
            return

        q = session.query(model.Product)
        q = q.filter(model.Product.upc == upc)
        product = q.first()
        if not product:
            return

        # Remove associated ProductPrice records.
        product.regular_price = None
        product.current_price = None
        q = session.query(model.ProductPrice)
        q = q.filter(model.ProductPrice.product == product)
        q.delete(synchronize_session=False)

        # Remove associated ProductCost records.
        q = session.query(model.ProductCost)
        q = q.filter(model.ProductCost.product == product)
        q.delete(synchronize_session=False)

        session.delete(product)

    def process_pos_chg(self, row, session):
        upc = row['F01']
        if not upc.isdigit():
            log.warning("non-numeric upc not supported: {0}".format(repr(upc)))
            return

        product = self.get_product(upc, session)

        subdepartment = None
        number = self.int_(row['F04'])
        if number:
            q = session.query(model.Subdepartment)
            q = q.filter(model.Subdepartment.number == number)
            subdepartment = q.first()
            if not subdepartment:
                subdepartment = model.Subdepartment(number=number)
                session.add(subdepartment)

        product.subdepartment = subdepartment
        if subdepartment:
            product.department = subdepartment.department
        else:
            product.department = None

        if row['F106'] == '1':  # follow subdept
            if subdepartment:
                sms_session = SmsSession()
                sms_subdept = sms_session.query(locsms.Subdepartment).get(subdepartment.number)
                product.weighed = bool(sms_subdept) and sms_subdept.F82 == '1' # scalable
                sms_session.close()
            else:
                product.weighed = False
        else:
            product.weighed = row['F82'] == '1' # scalable

        product.not_for_sale = row['F86'] == '1'
        return product

    def process_pos_del(self, row, session):
        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return

        upc = GPC(upc, calc_check_digit='upc')

        if row['F1000'] != 'PAL':
            log.warning("target {0} not supported for upc {1}".format(
                    repr(row['F1000']), repr(upc)))
            return

        q = session.query(model.Product)
        q = q.filter(model.Product.upc == upc)
        product = q.first()
        if not product:
            return

        product.subdepartment = None

    def process_alt_chg(self, row, session):
        if row['F1000'] != 'PAL':
            log.debug("target {0} not supported for upc {1}".format(
                    repr(row['F1000']), repr(row['F01'])))
            return

        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return

        product = self.get_product(upc, session)
        if not product:
            log.warning("product not found: {0}".format(repr(upc)))
            return

        sms_code = row['F154']
        is_main = row['F1898'] == '1'
        if sms_code not in product.codes:
            if is_main:
                product.codes.insert(0, sms_code)
            else:
                product.codes.append(sms_code)
        elif is_main and product.code != sms_code:
            for i, code in enumerate(product.codes):
                if code == sms_code:
                    code = product.codes.pop(i)
                    product.codes.insert(0, code)
                    break

        return product

    def process_alt_del(self, row, session):
        if row['F1000'] != 'PAL':
            log.debug("target {0} not supported for upc {1}".format(
                    repr(row['F1000']), repr(row['F01'])))
            return

        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return

        product = self.get_product(upc, session)
        if not product:
            log.warning("product not found: {0}".format(repr(upc)))
            return

        code = row['F154']
        if code in product.codes:
            product.codes.remove(code)

    def process_price_chg(self, row, session):
        """
        Process a price change.
        """
        upc = row[u'F01']
        if not upc.isdigit():
            log.warning(u"invalid upc: {0}".format(repr(upc)))
            return

        if row[u'F1000'] != u'PAL':
            log.debug(u"ignoring price change for target {0} for upc: {1}".format(
                    repr(row[u'F1000']), repr(row[u'F01'])))
            return

        product = self.get_product(upc, session)
        sms_price = self.get_price_object(row)
        prices = session.query(model.ProductPrice)

        # Create / update Regular price.
        if sms_price.F30 or sms_price.F140:
            starts = self.utc_timestamp(sms_price.F35, sms_price.F36)
            q = prices.filter(model.ProductPrice.product == product)
            q = q.filter(model.ProductPrice.type == enum.PRICE_TYPE_REGULAR)
            q = q.filter(model.ProductPrice.level == sms_price.F126)
            q = q.filter(model.ProductPrice.starts == starts)
            price = q.first()
            if not price:
                price = model.ProductPrice()
                price.type = enum.PRICE_TYPE_REGULAR
                price.level = sms_price.F126
                price.starts = starts
                product.prices.append(price)
            self.copy_reg_price(sms_price, price)
            if price.level == 1:
                product.regular_price = price if self.price_is_current(price) else None

        # Create / update TPR price.
        tpr_price = None
        if sms_price.F181 or sms_price.F1186:
            starts = self.utc_timestamp(sms_price.F183, None)
            q = prices.filter(model.ProductPrice.product == product)
            q = q.filter(model.ProductPrice.type == enum.PRICE_TYPE_TPR)
            q = q.filter(model.ProductPrice.level == sms_price.F126)
            q = q.filter(model.ProductPrice.starts == starts)
            price = q.first()
            if not price:
                price = model.ProductPrice()
                price.type = enum.PRICE_TYPE_TPR
                price.level = sms_price.F126
                price.starts = starts
                product.prices.append(price)
            self.copy_tpr_price(sms_price, price)
            tpr_price = price

        # Create / update Sale price.
        sale_price = None
        sale_is_current = False
        if sms_price.F136 or sms_price.F139:
            starts = self.utc_timestamp(sms_price.F137, sms_price.F144)
            q = prices.filter(model.ProductPrice.product == product)
            q = q.filter(model.ProductPrice.type == enum.PRICE_TYPE_SALE)
            q = q.filter(model.ProductPrice.level == sms_price.F126)
            q = q.filter(model.ProductPrice.starts == starts)
            price = q.first()
            if not price:
                price = model.ProductPrice()
                price.type = enum.PRICE_TYPE_SALE
                price.level = sms_price.F126
                price.starts = starts
                product.prices.append(price)
            self.copy_sale_price(sms_price, price)
            sale_price = price

        # Assign "current" price (Sale/TPR/None).
        if sale_price:
            if self.price_is_current(sale_price):
                product.current_price = sale_price
            elif tpr_price:
                if self.price_is_current(tpr_price):
                    product.current_price = tpr_price
                else:
                    product.current_price = None
            else:
                product.current_price = None
        elif tpr_price:
            if self.price_is_current(tpr_price):
                product.current_price = tpr_price
            else:
                product.current_price = None
        else:
            product.current_price = None

        return product

    def process_price_del(self, row, session):
        """
        Process a price deletion.
        """
        upc = row[u'F01']
        if not upc.isdigit():
            log.warning(u"invalid upc: {0}".format(repr(upc)))
            return

        if row[u'F1000'] != u'PAL':
            log.debug(u"ignoring deletion on target {0} for upc: {1}".format(
                    repr(row[u'F1000']), repr(row[u'F01'])))
            return

        upc = GPC(upc, calc_check_digit=u'upc')
        product = get_product_by_upc(session, upc)
        if not product:
            return

        prices = session.query(model.ProductPrice)\
            .filter(model.ProductPrice.product == product)\
            .filter(model.ProductPrice.level == self.int_(row[u'F126']))

        for price in prices:
            log.debug(u"deleting level {0} price for upc: {1}".format(
                    price.level, repr(row[u'F01'])))
            if product.regular_price is price:
                product.regular_price = None
            if product.current_price is price:
                product.current_price = None
            session.delete(price)

    def process_cost_chg(self, row, session):
        if row['F1000'] != 'PAL':
            log.warning("target {0} not supported for upc {1}".format(
                    repr(row['F1000']), repr(row['F01'])))
            return

        if row['F1184'] != 'CASE':
            log.warning("buying format {0} not supported: {1}".format(
                    repr(row['F1184']), repr(row['F01'])))
            return

        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return

        product = self.get_product(upc, session)
        upc = product.upc

        q = session.query(model.Vendor)
        q = q.filter(model.Vendor.id == row['F27'])
        vendor = q.first()
        if not vendor:
            vendor = model.Vendor(id=row['F27'])
            session.add(vendor)

        q = session.query(model.ProductCost)
        q = q.filter(model.ProductCost.product == product)
        q = q.filter(model.ProductCost.vendor == vendor)
        cost = q.first()
        if not cost:
            cost = model.ProductCost()
            cost.vendor = vendor
            product.costs.append(cost)

        cost.code = row['F26'].strip() or None
        try:
            cost.case_size = self.int_(row['F19'])
        except ValueError:
            log.warning("case size for product {0} is not integer, but will be "
                        "coerced: {1}".format(repr(upc), repr(row['F19'])))
            cost.case_size = int(Decimal(row['F19']))
        cost.case_cost = self.decimal(row['F38'])
        cost.unit_cost = self.decimal(row['F1140'])

        if row['F90'] == '1' and cost.preference != 1:
            cost = product.costs.pop(cost.preference - 1)
            product.costs.insert(0, cost)

        return cost

    def process_cost_del(self, row, session):
        upc = row['F01']
        if not upc.isdigit():
            log.warning("invalid upc: {0}".format(repr(upc)))
            return

        upc = GPC(upc, calc_check_digit='upc')

        if row['F1000'] != 'PAL':
            log.warning("target {0} not supported for {1}".format(
                    repr(row['F1000']), repr(upc)))
            return

        if row['F1184'] != 'CASE':
            log.warning("buying format {0} not supported for {1}".format(
                    repr(row['F1184']), repr(upc)))
            return

        q = session.query(model.Product)
        q = q.filter(model.Product.upc == upc)
        product = q.first()
        if not product:
            return

        q = session.query(model.Vendor)
        q = q.filter(model.Vendor.id == row['F27'])
        vendor = q.first()
        if not vendor:
            return

        q = session.query(model.ProductCost)
        q = q.filter(model.ProductCost.product == product)
        q = q.filter(model.ProductCost.vendor == vendor)
        cost = q.first()
        if not cost:
            return

        product.costs.remove(cost)
        session.delete(cost)

    def process_clt_chg(self, row, session):
        customer_id = row[u'F1148'].strip()
        assert customer_id
        try:
            customer = session.query(model.Customer)\
                .filter(model.Customer.id == customer_id)\
                .one()
        except NoResultFound:
            person = model.Person()
            person.first_name = row['F1149'].strip()
            person.last_name = row['F1150']
            customer = model.Customer(id=customer_id)
            customer.people.append(person)
            session.add(customer)

        customer.name = row['F1155']

        # Phone Number
        phones = filter(lambda p: p.type == 'Voice', customer.phones)
        number = row['F1172'].strip()
        if number:
            if phones:
                phone = phones[0]
                if phone.number != number:
                    phone.number = number
            else:
                customer.add_phone_number(number, 'Voice')
        else:
            for phone in phones:
                customer.phones.remove(phone)

        # Phone Number 2
        phones = filter(lambda p: p.type == 'Voice', customer.phones)
        number = row['F1173'].strip()
        if number:
            if len(phones) > 1:
                phone = phones[1]
                if phone.number != number:
                    phone.number = number
            else:
                customer.add_phone_number(number, 'Voice')
        else:
            for phone in phones[1:]:
                customer.phones.remove(phone)

        # Email Address
        email_address = row[u'F1573']
        if email_address:
            if customer.emails:
                assert len(customer.emails) == 1
                customer.emails[0].address = email_address
            else:
                customer.emails.append(model.CustomerEmailAddress(address=email_address))
        elif customer.emails:
            assert len(customer.emails) == 1
            session.delete(customer.emails.pop(0))

        def fetch_group(group_id):
            try:
                group = session.query(model.CustomerGroup)\
                    .filter(model.CustomerGroup.id == group_id)\
                    .one()
            except NoResultFound:
                group = model.CustomerGroup(id=group_id)
                session.add(group)
            return group

        # Groups
        group_ids = [row[u'F1154'].strip(), row[u'F1777'].strip()]
        group1, group2 = None, None
        if group_ids[0]:
            group1 = fetch_group(group_ids[0])
            if customer.groups:
                customer._groups[0].group = group1
            else:
                customer.groups.append(group1)
        if group_ids[1]:
            group2 = fetch_group(group_ids[1])
            if group1:
                if len(customer.groups) > 1:
                    assert len(customer.groups) == 2
                    customer._groups[1].group = group2
                else:
                    customer.groups.append(group2)
            else:
                if not customer.groups:
                    customer.groups.append(group2)
                elif len(customer.groups) == 1:
                    customer._groups[0].group = group2
                else:
                    assert len(customer.groups) == 2
                    customer._groups[0].group = group2
                    session.delete(customer._groups.pop(1))
        elif group1:
            if len(customer.groups) > 1:
                assert len(customer.groups) == 2
                session.delete(customer._groups.pop(1))
        elif customer.groups:
            for i in range(len(customer.groups)):
                session.delete(customer._groups.pop(0))

        return customer

    def process_clt_del(self, row, session):
        q = session.query(model.Customer)
        q = q.filter(model.Customer.id == row['F1148'])
        customer = q.first()
        if not customer:
            return

        session.delete(customer)

    def process_clk_chg(self, row, session):
        """
        Process a ``CLK_CHG`` row, i.e. add/update an employee.
        """
        employee_id = self.int_(row['F1126'])
        if not employee_id:
            log.error("CLK_CHG row has no F1126 value: {0}".format(repr(row)))
            return

        q = session.query(model.Employee)
        q = q.filter(model.Employee.id == employee_id)
        employee = q.first()
        if employee:
            while employee.emails:
                employee.emails.pop()
            while employee.phones:
                employee.phones.pop()
        else:
            person = model.Person()
            employee = model.Employee(id=employee_id)
            employee.person = person
            session.add(employee)

        # Update employee.
        employee.first_name = row['F1143'].strip()
        employee.last_name = row['F1144']
        employee.display_name = row['F1127']

        # Maybe load/create customer record.
        if row['F1148']:
            customer = session.query(model.Customer).filter_by(id=row['F1148']).first()
            if not customer:
                customer = model.Customer(id=row['F1148'])
                session.add(customer)

            # Make sure employee is attached to customer.
            if employee.person not in customer.people:
                customer.people.append(employee.person)

            # If employee is the only person attached to customer, update name.
            if len(customer.people) == 1:
                customer.name = employee.person.display_name

        status = (row['F1552'] or '').upper()
        if status == 'ACTV':
            employee.status = enum.EMPLOYEE_STATUS_CURRENT
        elif status == 'FRMR':
            employee.status = enum.EMPLOYEE_STATUS_FORMER
        else:
            employee.status = None

        if row['F1563']:
            employee.add_phone_number(row['F1563'])
        if row['F1564']:
            employee.add_phone_number(row['F1564'])
        if row['F1571']:
            employee.add_email_address(row['F1571'])

        return employee

    def process_clk_del(self, row, session):
        q = session.query(model.Employee)
        q = q.filter(model.Employee.id == row['F1126'])
        employee = q.first()
        if not employee:
            return

        session.delete(employee)


class ProcessChanges(Action):
    """
    Processes "changes" CSV data which has been deployed from an SMS server
    instance, with the Deploy_Rattail option installed.

    This action is designed to be called by the file monitor.  An example
    configuration would be (note that some things are optional):

    .. code-block:: ini

       [rattail.filemon]
       monitor = sms_changes
       sms_changes.dirs = C:\Storeman\XchRattail\Changes
       sms_changes.actions = process
       sms_changes.action.process.class = rattail_locsms.changes:ProcessChanges
       sms_changes.action.process.kwarg.processor_factory = mymodule:MyChangesProcessor
       sms_changes.action.process.kwarg.backup_dir = C:\Storeman\XchRattail\Changes\Backup
    """

    def __call__(self, path, backup_dir='processed', processor_factory=None, lock_table=False):
        """
        Process a change file.

        Note that in reality, the only time processing will occur is when the
        final "CHANGES_END" file is seen.  I.e. this method will merely ignore
        all files until the "end" file, at which point all files deemed part of
        the current "batch" will be processed at once.

        If no custom ``processor_factory`` is provided, it will default to the
        :class:`ChangesProcessor` class.

        Once processing is complete, all files from the current batch will be
        moved to the ``backup_dir`` location.
        """
        if isinstance(lock_table, basestring):
            lock_table = lock_table.strip().lower() == 'true'

        match = change_pattern_generic.match(os.path.basename(path))
        if not match:
            log.warning("unexpected filename pattern: {0}".format(repr(path)))
            return

        # Ignore all incoming files until we see one indicating "batch" completion.
        if match.group('batch_name').upper() != 'CHANGES_END':
            return

        deploy_id = match.group('batch_id')

        changes = {}
        changes_dir = os.path.dirname(path)
        for fn in os.listdir(changes_dir):
            if fn.startswith(deploy_id):
                match = change_pattern.match(fn)
                if match:
                    changes[match.group('batch_name').upper()] = os.path.join(changes_dir, fn)

        if processor_factory is not None:
            log.debug("using custom changes processor factory: {0}".format(processor_factory))
            processor_factory = load_object(processor_factory)
        else:
            processor_factory = ChangesProcessor

        user = self.config.get('rattail_locsms', 'continuum.user')
        session = Session(continuum_user=user)
        processor = processor_factory(self.config)
        if lock_table:
            session.execute('LOCK TABLE change')
        processor.process_all_changes(session, changes)
        session.commit()
        session.close()

        if os.path.basename(backup_dir) == backup_dir:
            backup_dir = os.path.join(changes_dir, backup_dir)
        else:
            backup_dir = os.path.abspath(backup_dir)

        log.debug("backing up data files in dir: {0}".format(backup_dir))
        if not os.path.exists(backup_dir):
            os.makedirs(backup_dir)
        for fn in list(os.listdir(changes_dir)):
            if fn.startswith(deploy_id):
                overwriting_move(os.path.join(changes_dir, fn), backup_dir)
