# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Importing from LOC SMS
"""

from __future__ import unicode_literals

import datetime
import logging
from decimal import Decimal

from sqlalchemy.orm import joinedload, joinedload_all

from locsms.db import model as locsms

from rattail import enum
from rattail.db import model
from rattail.db import importing
from rattail.gpc import GPC
from rattail.db.util import maxlen
from rattail.time import localtime, make_utc


log = logging.getLogger(__name__)


def import_from_locsms(config, session, sms_session, models=None, providers=None,
                       warnings=False, progress=None):
    """
    Import pertinent data from a BRdata export file.
    """
    if not models:
        models = [
            'Store',
            'Employee',
            'CustomerGroup',
            'Customer',
            'Vendor',
            'Department',
            'Subdepartment',
            'Family',
            'ReportCode',
            'Product',
            'ProductCode',
            'ProductCost',
            ]
    if providers is None:
        providers = globals()

    for name in models:

        # Create the provider object.
        provider_name = '{0}Provider'.format(name)
        if hasattr(providers, 'get'):
            Provider = providers.get(provider_name)
        else:
            Provider = getattr(providers, provider_name, None)
        assert Provider, "No provider class found for model: {0}".format(repr(name))
        provider = Provider(config, session, sms_session=sms_session)

        # Import some data!
        data = provider.get_data(progress=progress)
        affected = provider.importer.import_data(data, provider.supported_fields, provider.key, progress=progress)
        log_func = log.warning if warnings and affected else log.info
        log_func("added or updated {0} {1} records".format(affected, name))


class SmsProvider(importing.QueryProvider):
    """
    Base class for all LOC SMS data providers.
    """

    def query(self):
        return self.sms_session.query(self.sms_model_class)

    @property
    def normalize_progress_message(self):
        return "Reading {0} data from LOC SMS".format(self.model_name)

    def make_utc(self, time):
        if time is None:
            return None
        return make_utc(localtime(self.config, time))

    def make_time(self, date, time):
        """
        Combine a date and time object to create a proper datetime.
        """
        if date is None:
            return None
        if time is None:
            time = datetime.time(0) # assume midnight
        else:
            time = datetime.datetime.strptime(time, '%H%M').time()
        return datetime.datetime.combine(date, time)


class StoreProvider(SmsProvider):
    """
    Provider for store data from LOC SMS.
    """
    importer_class = importing.models.StoreImporter
    key = 'id'
    sms_model_class = locsms.Store
    supported_fields = [
        'id',
        'name',
        'phone_number',
        'fax_number',
        ]

    def normalize(self, store):
        return {
            'id': store.F1056,
            'name': store.F1531,
            'phone_number': store.F1536 or None,
            'fax_number': store.F1537 or None,
            }


class EmployeeProvider(SmsProvider):
    """
    Provider for employee data from LOC SMS.
    """
    importer_class = importing.models.EmployeeImporter
    key = 'id'
    sms_model_class = locsms.Operator
    supported_fields = [
        'id',
        'customer_id',
        'first_name',
        'last_name',
        'display_name',
        'status',
        'phone_number',
        'phone_number_2',
        'email_address',
        ]
    status_map = {
        'ACTV': enum.EMPLOYEE_STATUS_CURRENT,
        'FRMR': enum.EMPLOYEE_STATUS_FORMER,
        }

    def normalize(self, employee):
        return {
            'id': employee.F1126,
            'customer_id': employee.F1148,
            'first_name': employee.F1143,
            'last_name': employee.F1144,
            'display_name': employee.F1127,
            'status': self.status_map.get((employee.F1552 or '').upper()),
            'phone_number': employee.F1563,
            'phone_number_2': employee.F1564,
            'email_address': employee.F1571,
            }


class CustomerGroupProvider(SmsProvider):
    """
    Provider for customer group data from LOC SMS.
    """
    importer_class = importing.models.CustomerGroupImporter
    key = 'id'
    sms_model_class = locsms.CustomerGroup
    supported_fields = [
        'id',
        'name',
        ]

    def normalize(self, group):
        return {
            'id': group.F1154 or None,
            'name': group.F1268,
            }


class CustomerProvider(SmsProvider):
    """
    Provider for customer data from LOC SMS.
    """
    importer_class = importing.models.CustomerImporter
    key = 'id'
    sms_model_class = locsms.Customer
    supported_fields = [
        'id',
        'name',
        'first_name',
        'last_name',
        'phone_number',
        'phone_number_2',
        'email_address',
        'group_id',
        'group_id_2',
        ]

    def normalize(self, customer):
        return {
            'id': customer.F1148 or None,
            'name': customer.F1155,
            'first_name': customer.F1149,
            'last_name': customer.F1150,
            'phone_number': customer.F1172,
            'phone_number_2': customer.F1173,
            'email_address': customer.F1573,
            'group_id': customer.F1154,
            'group_id_2': customer.F1777,
            }


class VendorProvider(SmsProvider):
    """
    Provider for vendor data from LOC SMS.
    """
    importer_class = importing.models.VendorImporter
    key = 'id'
    sms_model_class = locsms.Vendor
    supported_fields = [
        'id',
        'name',
        'special_discount',
        'phone_number',
        'phone_number_2',
        'fax_number',
        'fax_number_2',
        'contact_name',
        'contact_name_2',
        'email_address',
        ]
    maxlen_name = maxlen(model.Vendor.name)

    def normalize(self, vendor):

        # Warn about truncation until Rattail schema is addressed.
        name = vendor.F334 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("vendor name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        special_discount = vendor.F347
        if isinstance(special_discount, float):
            special_discount = Decimal(unicode(special_discount))

        return {
            'id': vendor.F27,
            'name': name,
            'special_discount': special_discount,
            'phone_number': vendor.F341,
            'phone_number_2': vendor.F1890,
            'fax_number': vendor.F342,
            'fax_number_2': vendor.F1891,
            'contact_name': vendor.F335,
            'contact_name_2': vendor.F1889,
            'email_address': vendor.F2603,
            }


class DepartmentProvider(SmsProvider):
    """
    Provider for department data from LOC SMS.
    """
    importer_class = importing.models.DepartmentImporter
    key = 'number'
    sms_model_class = locsms.Department
    supported_fields = [
        'number',
        'name',
        ]
    maxlen_name = maxlen(model.Department.name)

    def normalize(self, department):

        # Warn about truncation until Rattail schema is addressed.
        name = department.F238 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("department name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'number': department.F03 or None,
            'name': name,
            }


class SubdepartmentProvider(SmsProvider):
    """
    Provider for subdepartment data from LOC SMS.
    """
    importer_class = importing.models.SubdepartmentImporter
    key = 'number'
    sms_model_class = locsms.Subdepartment
    supported_fields = [
        'number',
        'name',
        'department_number',
        ]
    maxlen_name = maxlen(model.Subdepartment.name)

    def normalize(self, subdepartment):

        # Warn about truncation until Rattail schema is addressed.
        name = subdepartment.F1022 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("subdepartment name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'number': subdepartment.F04 or None,
            'name': name,
            'department_number': subdepartment.F03 or None,
            }


class FamilyProvider(SmsProvider):
    """
    Provider for family data from LOC SMS.
    """
    importer_class = importing.models.FamilyImporter
    key = 'code'
    sms_model_class = locsms.Family
    supported_fields = [
        'code',
        'name',
        ]
    maxlen_name = maxlen(model.Family.name)

    def normalize(self, family):

        # Warn about truncation until Rattail schema is addressed.
        name = family.F1040 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("family name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'code': family.F16 or None,
            'name': name,
            }


class ReportCodeProvider(SmsProvider):
    """
    Provider for report code data from LOC SMS.
    """
    importer_class = importing.models.ReportCodeImporter
    key = 'code'
    sms_model_class = locsms.ReportCode
    supported_fields = [
        'code',
        'name',
        ]
    maxlen_name = maxlen(model.ReportCode.name)

    def normalize(self, reportcode):

        # Warn about truncation until Rattail schema is addressed.
        name = reportcode.F1024 or ''
        if self.maxlen_name and len(name) > self.maxlen_name:
            log.warning("report code name is more than {} chars and will be truncated: {}".format(
                    self.maxlen_name, repr(name)))
            name = name[:self.maxlen_name]

        return {
            'code': reportcode.F18 or None,
            'name': name,
            }


class ProductProvider(SmsProvider):
    """
    Provider for product data from LOC SMS.
    """
    importer_class = importing.models.ProductImporter
    key = 'upc'
    sms_model_class = locsms.Product
    supported_fields = [
        'upc',
        'brand_name',
        'description',
        'size',
        'unit_of_measure',
        'not_for_sale',
        'department_number',
        'subdepartment_number',
        'family_code',
        'report_code',
        'regular_price_price',
        'regular_price_multiple',
        'regular_price_pack_price',
        'regular_price_pack_multiple',
        'regular_price_type',
        'regular_price_level',
        'regular_price_starts',
        'regular_price_ends',
        'sale_price_price',
        'sale_price_multiple',
        'sale_price_pack_price',
        'sale_price_pack_multiple',
        'sale_price_type',
        'sale_price_level',
        'sale_price_starts',
        'sale_price_ends',
        ]
    maxlen_description = maxlen(model.Product.description)
    maxlen_size = maxlen(model.Product.size)

    def query(self):
        return self.sms_session.query(locsms.Product)\
            .options(joinedload(locsms.Product.price))\
            .options(joinedload_all(
                locsms.Product.pos,
                locsms.POS.subdepartment,
                locsms.Subdepartment.department))

    def normalize(self, product):

        try:
            upc = GPC(product.F01, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.warning("SMS product has invalid UPC: {0}".format(repr(product.F01)))
            upc = None

        uom = enum.UNIT_OF_MEASURE_EACH
        pos = product.pos.get('PAL')
        if pos:
            if pos.F106 == '1':
                if pos.subdepartment and pos.subdepartment.F82 == '1':
                    uom = enum.UNIT_OF_MEASURE_POUND
            elif pos.F82 == '1':
                uom = enum.UNIT_OF_MEASURE_POUND

        # Warn about truncation until Rattail schema is addressed.
        description = product.F29 or ''
        if self.maxlen_description and len(description) > self.maxlen_description:
            log.warning("product description is more than {} chars and will be truncated: {}".format(
                    self.maxlen_description, repr(description)))
            description = description[:self.maxlen_description]

        # Warn about truncation until Rattail schema is addressed.
        size = product.F22 or ''
        if self.maxlen_size and len(size) > self.maxlen_size:
            log.warning("product size is more than {} chars and will be truncated: {}".format(
                    self.maxlen_size, repr(size)))
            size = size[:self.maxlen_size]

        # Determine if sale price is currently active; if it is not then we
        # will declare None for all sale fields.
        price = product.price
        sale_starts = self.make_utc(price.sale_start if price else None)
        sale_ends = self.make_utc(price.sale_stop if price else None)
        sale_price = price.F136 if price else None
        if sale_price:
            active = False
            if sale_starts and sale_ends:
                if sale_starts <= self.now <= sale_ends:
                    active = True
            elif sale_starts:
                if sale_starts <= self.now:
                    active = True
            elif sale_ends:
                if self.now <= sale_ends:
                    active = True
            else:
                active = True
            if not active:
                sale_price = None

        return {
            'upc': upc,
            'brand_name': product.F155,
            'description': description,
            'size': size,
            'unit_of_measure': uom,
            'not_for_sale': pos.F86 == '1' if pos else False,

            'department_number': product.department.F03 if product.department else None,
            'subdepartment_number': product.subdepartment.F04 if product.subdepartment else None,
            'family_code': product.F16,
            'report_code': product.F18,

            'regular_price_price': price.F30 if price else None,
            'regular_price_multiple': (price.F31 or 1) if price else None,
            'regular_price_pack_price': price.F140 if price and price.F140 else None,
            'regular_price_pack_multiple': (price.F142 or 1) if price and price.F140 else None,
            'regular_price_type': enum.PRICE_TYPE_REGULAR if price else None,
            'regular_price_level': price.F126 if price else None,
            'regular_price_starts': self.make_utc(self.make_time(price.F35, price.F36)) if price else None,
            'regular_price_ends': self.make_utc(self.make_time(price.F129, price.F130)) if price else None,

            'sale_price_price': sale_price,
            'sale_price_multiple': (price.F135 or 1) if sale_price else None,
            'sale_price_pack_price': price.F139 if sale_price else None,
            'sale_price_pack_multiple': price.F143 if sale_price else None,
            'sale_price_type': enum.PRICE_TYPE_SALE if sale_price else None,
            'sale_price_level': price.F126 if sale_price else None,
            'sale_price_starts': sale_starts if sale_price else None,
            'sale_price_ends': sale_ends if sale_price else None,
            }


class ProductCodeProvider(SmsProvider):
    """
    Provider for product code data from LOC SMS.
    """
    importer_class = importing.models.ProductCodeImporter
    key = ('product_upc', 'code')
    sms_model_class = locsms.AlternateCode
    supported_fields = [
        'product_upc',
        'ordinal',
        'code',
        ]

    def setup(self, progress=None):
        self.product_upc = None
        self.ordinal = None

    def query(self):
        return self.sms_session.query(locsms.AlternateCode)\
            .filter(locsms.AlternateCode.F1000 == 'PAL')\
            .order_by(locsms.AlternateCode.F01,
                      locsms.AlternateCode.F1898.desc(),
                      locsms.AlternateCode.F154)

    def normalize(self, code):
        try:
            upc = GPC(code.F01, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.warning("SMS code has invalid UPC: {0}".format(repr(code.F01)))
            upc = None

        if self.product_upc and upc == self.product_upc:
            self.ordinal += 1
        else:
            self.product_upc = upc
            self.ordinal = 1

        return {
            'product_upc': upc,
            'ordinal': self.ordinal,
            'code': code.F154,
            }


class ProductCostProvider(SmsProvider):
    """
    Provider for product cost data from LOC SMS.
    """
    importer_class = importing.models.ProductCostImporter
    key = ('product_upc', 'vendor_id')
    sms_model_class = locsms.Cost
    supported_fields = [
        'product_upc',
        'vendor_id',
        'code',
        'case_size',
        'case_cost',
        'unit_cost',
        'preference',
        ]

    def setup(self, progress):
        self.product_upc = None
        self.preference = None

    def query(self):
        return self.sms_session.query(locsms.Cost)\
            .filter(locsms.Cost.F1000 == 'PAL')\
            .filter(locsms.Cost.F1184 == 'CASE')\
            .order_by(locsms.Cost.F01, locsms.Cost.F90.desc(), locsms.Cost.F27)

    def normalize(self, cost):
        try:
            upc = GPC(cost.F01, calc_check_digit='upc')
        except (TypeError, ValueError):
            log.warning("SMS cost has invalid UPC: {0}".format(repr(cost.F01)))
            upc = None

        if upc and cost.F27:

            if self.product_upc and upc == self.product_upc:
                self.preference += 1
            else:
                self.product_upc = upc
                self.preference = 1

            # For some reason these values may exist as floats in some cases, I
            # think due to FreeTDS version differences?  In any event we want
            # to ensure decimal values.
            F38 = cost.F38
            if isinstance(F38, float):
                F38 = Decimal('{0:0.4f}'.format(F38))
            F1140 = cost.F1140
            if isinstance(F1140, float):
                F1140 = Decimal('{0:0.4f}'.format(F1140))

            case_size = (int(cost.F19) or 1) if cost.F19 else 1
            unit_cost = F38 / case_size if F38 else F1140
            if unit_cost:
                # Match scale of ProductCost field to aid data comparison.
                unit_cost = unit_cost.quantize(Decimal('.00001'))

            return {
                'product_upc': upc,
                'vendor_id': cost.F27,
                'code': cost.F26,
                'case_size': case_size,
                'case_cost': F38,
                'unit_cost': unit_cost,
                'preference': self.preference,
                }
